
# from pathlib import Path
# from pyhanko.sign.validation.pdf_embedded import EmbeddedPdfSignature
# from pyhanko_certvalidator import ValidationContext
# from pyhanko.keys import load_certs_from_pemder_data
# from pyhanko.pdf_utils.reader import PdfFileReader
# from pyhanko.sign.validation import validate_pdf_signature
# from pyhanko.sign.validation.status import PdfSignatureStatus
# from pypomes_core import file_get_data, exc_format, env_get_str, APP_PREFIX

class CryptoPdf:
    """
    Python code to extract cryptographic data from a digitally signed PDF file.

    instance attributes:

    """



from .crypto_pkcs7 import (
    CryptoPkcs7,
)
from .crypto_pomes import (
    CRYPTO_DEFAULT_HASH_ALGORITHM,
    crypto_validate_p7s, crypto_validate_pdf,
    crypto_compute_hash, crypto_generate_rsa_keys,
)

__all__ = [
    # crypto_pkcs7
    "CryptoPkcs7",
    # crypto_pomes
    "CRYPTO_DEFAULT_HASH_ALGORITHM",
    "crypto_validate_p7s", "crypto_validate_pdf",
    "crypto_compute_hash", "crypto_generate_rsa_keys",
]

from importlib.metadata import version
__version__ = version("pypomes_crypto")
__version_info__ = tuple(int(i) for i in __version__.split(".") if i.isdigit())



import hashlib
import pickle
import sys
from asn1crypto.x509 import Certificate
from collections.abc import Iterable
from contextlib import suppress
from Crypto.Hash import SHA256
from Crypto.Hash.SHA256 import SHA256Hash
from Crypto.PublicKey.RSA import import_key, RsaKey
from Crypto.Signature import pkcs1_15
from Crypto.Signature.pkcs1_15 import PKCS115_SigScheme
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from io import BytesIO
from pathlib import Path

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from pyhanko.sign.validation.pdf_embedded import EmbeddedPdfSignature
from pyhanko_certvalidator import ValidationContext
from pyhanko.keys import load_certs_from_pemder_data
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.sign.validation import validate_pdf_signature
from pyhanko.sign.validation.status import PdfSignatureStatus
from pypomes_core import file_get_data, exc_format, env_get_str, APP_PREFIX
from typing import Any, Final

from crypto_pkcs7 import CryptoPkcs7

CRYPTO_DEFAULT_HASH_ALGORITHM: Final[str] = \
    env_get_str(key=f"{APP_PREFIX}_CRYPTO_DEFAULT_HASH_ALGORITHM",
                def_value="sha256")


def crypto_validate_p7s(errors: list[str],
                        p7s_file: Path | str | bytes,
                        p7s_payload: str | bytes = None) -> bool:
    """
    Validate the digital signature of a PKCS#7 file.

    If a *list* is provided in *errors*, the following inconsistencies are reported:
        - The digital signature is invalid
        - Error from CryptoPkcs7 instantiation

    :param errors: incidental error messages
    :param p7s_file: a p7s file path, or the bytes thereof
    :param p7s_payload: a payload file path, or the bytes thereof
    :return: 'True' if the input data are consistent, 'False' otherwise
    """
    # instantiate the return variable
    result: bool = False

    # instantiate the PKCS7 object
    pkcs7: CryptoPkcs7 | None = None
    try:
        pkcs7 = CryptoPkcs7(p7s_file=p7s_file,
                            p7s_payload=p7s_payload)
    except Exception as e:
        if isinstance(errors, list):
            errors.append(exc_format(exc=e,
                                     exc_info=sys.exc_info()))

    # was the PKCS7 object instanciated ?
    if pkcs7:
        # yes, verify the signature
        try:
            # noinspection PyUnboundLocalVariable
            rsa_key: RsaKey = import_key(extern_key=pkcs7.public_bytes)
            sig_scheme: PKCS115_SigScheme = pkcs1_15.new(rsa_key=rsa_key)
            sha256_hash: SHA256Hash = SHA256.new(data=pkcs7.payload)
            # TODO @gtnunes: fix the verification process
            sig_scheme.verify(msg_hash=sha256_hash,
                              signature=pkcs7.signature)
            result = True
        except ValueError:
            if isinstance(errors, list):
                errors.append("The digital signature is invalid")
        except Exception as e:
            if isinstance(errors, list):
                errors.append(exc_format(exc=e,
                                         exc_info=sys.exc_info()))
    return result


def crypto_validate_pdf(errors: list[str] | None,
                        pdf_file: Path | str | bytes,
                        certs_file: Path | str | bytes = None) -> bool:
    """
    Validate the digital signature of a PDF file.

    If a *list* is provided in *errors*, the following inconsistencies are reported:
        - The file is not digitally signed
        - The digital signature is not valid
        - The certificate used has been revoked
        - The certificate used is not trusted
        - The signature block is not intact
        - A bad seed value found

    :param errors: incidental error messages
    :param pdf_file: a PDF file path, or the PDF file bytes
    :param certs_file: a path to a file containing a PEM/DER-encoded certificate chain, or the bytes thereof
    :return: 'True' if the input data are consistent, 'False' otherwise
    """
    # initialize the return variable
    result: bool = True

    # obtain the PDF reader
    pdf_bytes: bytes = file_get_data(file_data=pdf_file)
    pdf_reader: PdfFileReader = PdfFileReader(stream=BytesIO(initial_bytes=pdf_bytes))

    # obtain the validation context
    certs: Iterable[Certificate] | None = None
    if certs_file:
        certs_bytes: bytes = file_get_data(file_data=certs_file)
        if certs_bytes:
            certs = load_certs_from_pemder_data(cert_data_bytes=certs_bytes)
    validation_context: ValidationContext = ValidationContext(trust_roots=certs)

    # obtain the list of digital signatures
    signatures: list[EmbeddedPdfSignature] = pdf_reader.embedded_signatures

    # were signatures retrieved ?
    if signatures:
        # yes, traverse the list verifying them
        for signature in signatures:
            error: str | None = None
            status: PdfSignatureStatus = validate_pdf_signature(embedded_sig=signature,
                                                                signer_validation_context=validation_context)
            if status.revoked:
                error = "The certificate used has been revoked"
            elif not status.intact:
                error = "The signature block is not intact"
            elif not status.trusted and certs:
                error = "The certificate used is not trusted"
            elif not status.seed_value_ok:
                error = "A bad seed value found"
            elif not status.valid:
                error = "The digital signature is not valid"

            # has an error been flagged ?
            if error:
                # yes, report it
                if isinstance(errors, list):
                    errors.append(error)
                result = False
    else:
        # no, report the problem
        if isinstance(errors, list):
            errors.append("The file is not digitally signed")
        result = False

    return result


def crypto_compute_hash(msg: Path | str | bytes | Any,
                        alg: str = CRYPTO_DEFAULT_HASH_ALGORITHM) -> bytes:
    """
    Compute the hash of *msg*, using the algorithm specified in *alg*.

    Supported algorithms: *md5*, *blake2b*, *blake2s*, *sha1*, *sha224*, *sha256*, *sha384*,
    *sha512*, *sha3_224*, *sha3_256*, *sha3_384*, *sha3_512*, *shake_128*, *shake_256*.

    :param msg: the message to calculate the hash for, or a path to a file
    :param alg: the algorithm to use (defaults to an environment-defined value, or to 'sha256')
    :return: the hash value obtained, or 'None' if the hash could not be computed
    """
    # initialize the return variable
    result: bytes | None = None

    # instantiate the hasher
    hasher = hashlib.new(name=alg.lower())

    # what is the type of the message ?
    if isinstance(msg, bytes):
        # argument is type 'bytes'
        hasher.update(msg)
        result = hasher.digest()

    elif isinstance(msg, Path | str):
        # argument is a file path
        buf_size: int = 128 * 1024
        file_path: Path = Path(msg)
        with file_path.open(mode="rb") as f:
            file_bytes: bytes = f.read(buf_size)
            while file_bytes:
                hasher.update(file_bytes)
                file_bytes = f.read(buf_size)
        result = hasher.digest()

    else:
        # argument is unknown
        with suppress(Exception):
            data: bytes = pickle.dumps(obj=msg)
            if data:
                hasher.update(data)
                result = hasher.digest()

    return result


def crypto_generate_rsa_keys(key_size: int = 2048) -> (str, str):
    """
    Generate and return a matching pair of *RSA* private and public keys.

    The keys are serialized and *UTF-8* decoded to *str*.

    :param key_size: the key size (defaults to 2048 bytes)
    :return: a matching pair (private_key, public_key) of serialized RSA keys
    """
    # generate the private key
    priv_key: RSAPrivateKey = rsa.generate_private_key(public_exponent=65537,
                                                       key_size=key_size)
    priv_serialized: bytes = priv_key.private_bytes(encoding=serialization.Encoding.PEM,
                                                    format=serialization.PrivateFormat.PKCS8,
                                                    encryption_algorithm=serialization.NoEncryption())
    # obtain the matching public key
    pub_key: RSAPublicKey = priv_key.public_key()
    pub_serialized: bytes = pub_key.public_bytes(encoding=serialization.Encoding.PEM,
                                                 format=serialization.PublicFormat.SubjectPublicKeyInfo)
    # serialize and return the keys
    return priv_serialized.decode(), pub_serialized.decode()



from asn1crypto import cms
from datetime import datetime
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from cryptography.hazmat.primitives.serialization.pkcs7 import load_der_pkcs7_certificates
from cryptography.x509 import Certificate
from pathlib import Path
from pypomes_core import file_get_data

class CryptoPkcs7:
    """
    Python code to extract relevant data from a PKCS#7 signature file in DER or PEM format.
    """
    # instance attributes
    # p7s_bytes: bytes              - the PKCS#7 data
    # payload: bytes                - the payload (embedded or external)
    # payload_hash: bytes           - the payload hash
    # hash_algorithm: str           - the algorithm used to calculate the payload hash
    # signature: bytes              - the digital signature
    # signature_algorithm: str      - the algorithm used to generate the signature
    # signature_timestamp: datetime - the signature's timestamp
    # public_key: RSAPublicKey      - the RSA public key
    # public_bytes: bytes           - the serialized public key (in PEM format)
    # cert_chain: list[bytes]       - the serialized X509 certificate chain (in PEM format)

    def __init__(self,
                 p7s_file: Path | str | bytes,
                 p7s_payload: str | bytes = None) -> None:
        """
        Instantiate the PKCS#7 crypto class, and extract the relevant data.

        If a detached payload is specified in *p7s_payload*, it is validated against the
        payload's declared hash value.

        :param p7s_file: path for a PKCS#7 file in DER format, or the bytes thereof
        :param p7s_payload: a payload file path, or the bytes thereof
        :raises ValueError: if the payload is inconsistent with its declared hash value
        :raises InvalidSignature: if the digital signature is invalid
        """
        # obtain the PKCS#7 file data
        self.p7s_bytes: bytes = file_get_data(file_data=p7s_file)

        # extract the certificate chain and serialize it in PEM format
        certs: list[Certificate] = load_der_pkcs7_certificates(data=self.p7s_bytes)
        self.cert_chain: list[bytes] = [cert.public_bytes(encoding=Encoding.PEM)
                                        for cert in certs]

        #  extract the public key and serialize it in PEM format
        cert: Certificate = certs[0]
        # 'cert.public_key()' may return one of:
        #   DSAPublicKey, RSAPublicKey, EllipticCurvePublicKey,
        #   Ed25519PublicKey, Ed448PublicKey, X25519PublicKey, X448PublicKey
        self.public_key: RSAPublicKey = cert.public_key()
        self.public_bytes: bytes = self.public_key.public_bytes(encoding=Encoding.PEM,
                                                                format=PublicFormat.SubjectPublicKeyInfo)
        # extract the needed structures
        content_info: cms.ContentInfo = cms.ContentInfo.load(encoded_data=self.p7s_bytes)
        signed_data: cms.SignedData = content_info["content"]
        signer_info: cms.SignerInfo = signed_data["signer_infos"][0]

        # extract the needed components
        self.signature: bytes = signer_info["signature"].native
        self.signature_algorithm: str = signer_info["signature_algorithm"]["algorithm"].native
        self.hash_algorithm: str = signer_info["digest_algorithm"]["algorithm"].native

        signed_attrs = signer_info["signed_attrs"]
        for signed_attr in signed_attrs:
            match signed_attr["type"].native:
                case "message_digest":
                    self.payload_hash: bytes = signed_attr["values"][0].native
                case "signing_time":
                    self.signature_timestamp: datetime = signed_attr["values"][0].native

        # has a detached payload been specified ?
        if p7s_payload:
            # yes, load it
            self.payload: bytes = file_get_data(file_data=p7s_payload)
        else:
            # no, extract the embedded payload
            self.payload: bytes = signed_data["encap_content_info"]["content"].native

    def verify_hash(self) -> bool:
        """
        Verify whether the declared hash value effectively matches the payload.

        :return: 'True' if the hash value matches the payload, 'False' otherwise
        """
        from crypto_pomes import crypto_compute_hash
        effective_hash: bytes = crypto_compute_hash(msg=self.payload,
                                                    alg=self.hash_algorithm)
        return effective_hash == self.payload_hash

    def verify_signature(self) -> bool:
        """
        Verify whether the digital signature is valid.

        :return: 'True' if the digital signature is valid, 'False' otherwise
        """
        # verify the digital signature
        result: bool
        try:
            self.public_key.verify(signature=self.signature,
                                   data=self.payload,
                                   padding=PKCS1v15(),
                                   algorithm=SHA256())
            result = True
        except InvalidSignature:
            result = False

        return result
